/**
 * @file boost/rubber/range.hpp
 * @brief Type erasure for Mirror's compile-time ranges
 *
 *  Copyright 2008-2011 Matus Chochlik. Distributed under the Boost
 *  Software License, Version 1.0. (See accompanying file
 *  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef BOOST_RUBBER_RANGE_1103151301_HPP
#define BOOST_RUBBER_RANGE_1103151301_HPP

#include <boost/rubber/config.hpp>
#include <boost/rubber/wrap.hpp>

#include <boost/mirror/meta_prog/range.hpp>

#include <vector>
#include <functional>

BOOST_RUBBER_NAMESPACE_BEGIN

struct iteration_info
{
        bool is_first;
        bool is_last;
};

template <typename TypeErased>
class range
{
private:
        typedef typename aux::return_value<TypeErased>::type
        (*member_getter)(void);

        const std::vector<member_getter>& getters;
        typename std::vector<member_getter>::const_iterator pos;
public:
        range(const std::vector<member_getter>& _getters)
         : getters(_getters)
         , pos(getters.begin())
        { }

        inline bool empty(void) const
        {
                return getters.end() == pos;
        }

        inline size_t size(void) const
        {
                return getters.end() - pos;
        }

        inline typename aux::return_value<TypeErased>::type front(void) const
        {
                assert(!empty());
                return (*pos)();
        }

        inline typename aux::return_value<TypeErased>::type at(size_t idx) const
        {
                assert(idx < size());
                return (*(pos + idx))();
        }

        inline void step_front(void)
        {
                assert(!empty());
                ++pos;
        }

        void for_each(const std::function<void(const TypeErased&)>& func) const
        {
                if(!empty())
                {
                        range tmp = *this;
                        while(!tmp.empty())
                        {
                                func(tmp.front());
                                tmp.step_front();
                        }
                }
        }

        void for_each_ii(
                const std::function<
                        void(const TypeErased&, bool, bool)
                >& func
        ) const
        {
                if(!empty())
                {
                        auto tmp_pos = pos;
                        auto tmp_get = *tmp_pos;
                        ++tmp_pos;
                        struct iteration_info iter_info = {
                                true,
                                tmp_pos == getters.end()
                        };
                        func(tmp_get(), iter_info);
                        iter_info.is_first = false;
                        while(tmp_pos != getters.end())
                        {
                                tmp_get = *tmp_pos;
                                ++tmp_pos;
                                iter_info.is_last = tmp_pos == getters.end();
                                func(tmp_get(), iter_info);
                        }
                }
        }
};

template <class TypeErased>
class range_holder
{
private:
        typedef typename aux::return_value<TypeErased>::type
        (*getter)(void);

        std::vector<getter> _getters;

        template <typename ... MetaObjects>
        static std::vector<getter> init_getters(
                boost::mirror::mp::range<MetaObjects...>
        )
        {
                return {&wrap_shaped<MetaObjects, TypeErased>...};
        }
public:
        template <template <class> class Getter, class MetaObject>
        range_holder(Getter<MetaObject>)
         : _getters(
                init_getters(
                        typename boost::mirror::mp::as_range<
                                Getter<MetaObject>
                        >::type()
                )
        )
        { }

        range_holder(boost::mirror::mp::empty_range)
        { }

        range<TypeErased> all(void) const
        {
                return range<TypeErased>(_getters);
        }
};

BOOST_RUBBER_NAMESPACE_END

#endif //include guard

